import java.util.ArrayList;
import java.util.List;

public class ArrayCopyTest {

    public static interface Copier {

        void arrayCopy(int[] src, int srcPos, int[] dst, int dstPos, int length);

    }

    public static class SystemCopier implements Copier {

        public void arrayCopy(int[] src, int srcPos, int[] dst, int dstPos, int length) {
            System.arraycopy(src, srcPos, dst, dstPos, src.length);
        }

    }

    public static class NaiveCopier implements Copier {

        public void arrayCopy(int[] src, int srcPos, int[] dest, int destPos, int length) {
            if (srcPos > destPos) {
                for (int i = 0; i < length; i++) {
                    dest[destPos + i] = src[srcPos + i];
                }
            } else {
                for (int i = length - 1; i >= 0; i--) {
                    dest[destPos + i] = src[srcPos + i];
                }
            }
        }
    }

    public static class SmartCopier implements Copier {

        public void arrayCopy(int[] src, int srcPos, int[] dest, int destPos, int length) {
            if (length > 20) {
                System.arraycopy(src, srcPos, dest, destPos, length);
                return;
            }
            if (srcPos > destPos) {
                for (int i = 0; i < length; i++) {
                    dest[destPos + i] = src[srcPos + i];
                }
            } else {
                for (int i = length - 1; i >= 0; i--) {
                    dest[destPos + i] = src[srcPos + i];
                }
            }
        }
    }

    private static final int ITERATIONS = 10;

    private static final int CYCLES = 10 * 1000 * 1000;

    private static final int MAX_SIZE = 32;

    public static void main(String[] args) {

        Copier copier = null;
        String type = "system";

        if (args.length >= 1) {
            type = args[0];
        }

        if (type.equalsIgnoreCase("system")) {
            copier = new SystemCopier();
        }
        if (type.equalsIgnoreCase("naive")) {
            copier = new NaiveCopier();
        }
        if (type.equalsIgnoreCase("smart")) {
            copier = new SmartCopier();
        }

//        System.out.println("Using " + type + " copier");


        int[] warmSrc = new int[1];
        int[] warmDst = new int[1];
        for (int c = 0; c < CYCLES * ITERATIONS; c++) {
            copier.arrayCopy(warmSrc, 0, warmDst, 0, warmSrc.length);
        }

        for (int size = 0; size <= MAX_SIZE; size++) {
            System.out.print(size + " ");

            List<Long> results = new ArrayList<Long>();
            for (int i = 0; i < ITERATIONS; i++) {
                int[] src = new int[size];
                int[] dst = new int[size];

                long time1 = System.nanoTime();
                for (int c = 0; c < CYCLES; c++) {
                    copier.arrayCopy(src, 0, dst, 0, src.length);
                }
                long time2 = System.nanoTime();
                System.out.print((time2 - time1) + " ");
                results.add(time2-time1);
            }

            System.out.print(mean(results) + " ");
            System.out.print(stdev(results) + " ");

            System.out.println();

        }


    }

    private static long stdev(List<Long> results) {
        long mean = mean(results);

        long result = 0L;
        for(Long l : results) {
            result += Math.pow(l-mean , 2);
        }

        return (long)Math.sqrt(result)/(results.size()-1);
    }

    private static long mean(List<Long> results) {
        long result = 0L;
        for(Long l : results) {
            result += l;
        }
        return result/results.size();
    }

}
